from VMFFile import VMFFile
from VMFNode import getBounds
import copy
import numpy

OUTSIDE_MATERIAL = "DEV/DEV_BLENDMEASURE"
DOOR_DISTANCE_TOLERANCE = 16

def oppositeDirection(direction):
  if direction == "north":
    return "south"
  elif direction == "south":
    return "north"
  elif direction == "east":
    return "west"
  elif direction == "west":
    return "east"
  else:
    return None

# http://stackoverflow.com/questions/1401712/calculate-euclidean-distance-with-numpy
def euclideanDistance(x,y):   
  return numpy.sqrt(numpy.sum((x-y)**2))

def pointNearPlane(point,bounds):
  pointDistance = euclideanDistance(bounds[0],point)
  portalSize = euclideanDistance(bounds[0],bounds[1])
  return pointDistance < portalSize + DOOR_DISTANCE_TOLERANCE
    
def findPortalOnSolid(solid):
  side = None
  for child in solid.children:
    if child.name == "side": 
      if child.properties["material"] == OUTSIDE_MATERIAL :
        side = child
        break
  if side == None:
    print "ERROR: Every portal must have a solid having a side with the material "+OUTSIDE_MATERIAL +" marking the outside"
  return side.plane
    
def getTranslationVector(mapPortal,otherMapPortal):
  portalBounds = getBounds(mapPortal)
  portalSize = portalBounds[1]-portalBounds[0]
  otherPortalBounds = getBounds(otherMapPortal)
  otherPortalSize = otherPortalBounds[1]-otherPortalBounds[0]
  if numpy.array_equal(portalSize,otherPortalSize):
    vector = portalBounds - otherPortalBounds
    return vector[0]
  else:
    return None
    
class MapTile:
  def __init__(self):
    pass
    
  def fromfile(self,filename):
    self.map = VMFFile()
    self.map.fromfile(filename)
    self.bounds = self.map.root.GetBoundsRecurse()
    self.doors = self.analyzeDoors()
    self.filename = filename
    
  def deepcopy(self):
    deepcopy = MapTile()
    deepcopy.map = self.map.deepcopy()
    deepcopy.bounds = self.bounds
    deepcopy.doors = copy.deepcopy(self.doors)
    deepcopy.filename = self.filename
    return deepcopy
    
  def intersect(self, otherMap):
    upperBounds = numpy.append(self.bounds[1],otherMap.bounds[1]).reshape((2,3))
    upperBound = numpy.min(upperBounds, axis=0)
    lowerBounds = numpy.append(self.bounds[0],otherMap.bounds[0]).reshape((2,3))
    lowerBound = numpy.max(lowerBounds, axis=0)
    return numpy.array([lowerBound,upperBound])
  
  def translate(self, vector):
    self.map.root.TranslateRecurse(vector)
    self.bounds[0] += vector
    self.bounds[1] += vector
  
  def predicateOnDoor(self, node, name):
    found = False
    if node.name == name:
      for child in node.children:
        if child.name == "editor" and "visgroupid" in child.properties: 
          visgroupid = int(child.properties["visgroupid"])
          if visgroupid in self.map.visgroups and self.map.visgroups[visgroupid][:5] == "door_":
            found = True
            break
    return found
  
  def analyzeDoors(self):
    doors = dict({'north': [], 'east': [], 'south': [], 'west': []})
    doorNodes = self.map.root.FindRecurse(lambda node : self.predicateOnDoor(node,"solid"))
    # TODO: group entities and portals or support groups
    for doorNode in doorNodes:
      doors[self.map.visgroups[int(doorNode.FindRecurse(lambda child : child.name == "editor")[0].properties["visgroupid"])][5:]].append(doorNode.properties["id"])
    return doors # TODO: set self.doors instead
    
  def checkVisgroups(self, otherMap):
    for id, visgroup in self.map.visgroups.items():
      if not id in otherMap.map.visgroups:
        return False
      otherVisgroup = otherMap.map.visgroups[id]
      if not ( visgroup == otherVisgroup or otherVisgroup[:5] == "dummy" ): # TODO: remove dummy from all files
        return False
    return True
    
  def findConnections(self, otherMap):
    if not self.checkVisgroups(otherMap):
      print "ERROR: Other map has different visgroups"
      return None
    connections = []
    for direction, doorList in self.doors.items():
      if len(doorList) > 0:
        print "Base map has a door in direction",direction
        otherDoorList = otherMap.doors[oppositeDirection(direction)]
        if len(otherDoorList) > 0:
          print "Other map has a door in opposite direction"
          connections.append((direction,doorList,otherDoorList))
    print "Have",len(connections),"possible connections"
    return connections
    
  def findPortalsAndVector(self, otherMap, connection):
    mapPortal = self.findPortalOnSolidWithId(connection[1])
    otherMapPortal = otherMap.findPortalOnSolidWithId(connection[2])
    vector = getTranslationVector(mapPortal,otherMapPortal)
    return (vector, mapPortal, otherMapPortal)
    
  def append(self, otherMap, connection, vectors):
    otherMap = otherMap.deepcopy()
    return self.mend(otherMap, connection, vectors)
    
  def findPortalOnSolidWithId(self,id):
    solid = self.map.root.FindRecurse(lambda node : node.name == "solid" and node.properties["id"] == id)[0]
    return findPortalOnSolid(solid)
    
  def mend(self, otherMap, connection, vectors):
    vector, mapPortal, otherMapPortal = vectors
    
    if not otherMap == self:
      removed = otherMap.map.root.DeleteRecurse(lambda node : "classname" in node.properties and node.properties["classname"] == "info_player_start")
      print "Removed",removed,"info_player_start from other map"
      removed = otherMap.map.root.DeleteRecurse(lambda node : "classname" in node.properties and node.properties["classname"] == "prop_door_rotating" and pointNearPlane(node.origin,otherMapPortal))
      print "Removed",removed,"doors from other map"
    removed = otherMap.map.root.DeleteRecurse(lambda node : node.name == "solid" and node.properties["id"] == connection[2])
    print "Removed",removed,"solids from other map"
      
    entities = self.map.root.FindRecurse(lambda node : node.name == "entity" and not node.properties["classname"] == "func_detail" and pointNearPlane(node.origin,mapPortal))
    removed = 0
    for entity in entities:
      removed += entity.DeleteRecurse(lambda node : node.name == "editor")
    print "Removed",removed,"editor information from remaining entities in base map"
    removed = self.map.root.DeleteRecurse(lambda node : node.name == "solid" and node.properties["id"] == connection[1])
    print "Removed",removed,"solids from base map"

    if not otherMap == self:
      maxId = self.map.root.GetMaximumIdRecurse(0)
      otherMap.map.root.IncreaseIdRecurse(maxId)
      print "Increased IDs in other map by",maxId
      print "Translating other map with vector",vector,"..."
      otherMap.translate(vector)
      print "Adding other map..."
      self.map.root.AddOtherMap(otherMap.map.root)
      
    self.doors = self.analyzeDoors() # TODO: alternatively: increase ids and copy door information from other map
    print "Updated door info"
    
  def collidesWith(self, otherMap):
    intersection = self.intersect(otherMap)
    size = intersection[1]-intersection[0]
    return numpy.all(size > numpy.array([0,0,0]))
    
  def detectLoops(self):
    print "Detecting loops..."
    def printDistances(node,portal):
      otherPortal = findPortalOnSolid(node)
      #print otherPortal
      #print portal
      print getTranslationVector(portal, otherPortal)
      return False
    doorNodes = self.map.root.FindRecurse(lambda node : self.predicateOnDoor(node,"solid"))
    zeroVector = numpy.array([0,0,0])
    #print zeroVector
    for doorNode in doorNodes:
      portal = findPortalOnSolid(doorNode)
      otherDoorNodes = self.map.root.FindRecurse(lambda node : not node == doorNode and self.predicateOnDoor(node,"solid") and numpy.array_equal(getTranslationVector(portal,findPortalOnSolid(node)),zeroVector))
      if len(otherDoorNodes) == 1:
        otherDoorNode = otherDoorNodes[0]
        # TODO: the wrong door is removed
        self.mend(self,('',doorNode.properties["id"],otherDoorNode.properties["id"]), (zeroVector, portal, findPortalOnSolid(otherDoorNode)))
    
  def close(self):
    self.detectLoops()
    # TODO: sometimes not all remaining doors are removed
    removed = self.map.root.DeleteRecurse(lambda node : "classname" in node.properties and node.properties["classname"] == "prop_door_rotating" and self.predicateOnDoor(node,"entity"))
    print "Removed",removed,"doors to close map"
      
  def generateNavMeshScript(self):
    lines = []
    lines.append(["sv_cheats 1","z_debug 1","director_stop","nb_delete_all","nav_edit 1"])

    start = self.map.root.FindRecurse(lambda node : "classname" in node.properties and node.properties["classname"] == "info_null" and "targetname" in node.properties and node.properties["targetname"] == "start")
    if not len(start) == 2:
      print "ERROR: Need 2 corners for PLAYER_START nav mesh, got",len(start),"instead"
    else:
      lines.append(["nav_clear_selected_set","setpos " + start[0].GetOrigin() + "","setang 90 0 0"])
      lines.append(["nav_begin_area","setpos " + start[1].GetOrigin() + "","setang 90 0 0"])
      lines.append(["nav_end_area","nav_toggle_in_selected_set","mark PLAYER_START","nav_clear_selected_set","clear_attribute PLAYER_START"])
    
    finale = self.map.root.FindRecurse(lambda node : "classname" in node.properties and node.properties["classname"] == "info_null" and "targetname" in node.properties and node.properties["targetname"] == "finale")
    if not len(finale) == 2:
      print "ERROR: Need 2 corners for FINALE nav mesh, got",len(finale),"instead"
    else:
      lines.append(["nav_clear_selected_set","setpos " + finale[0].GetOrigin() + "","setang 90 0 0"])
      lines.append(["nav_begin_area","setpos " + finale[1].GetOrigin() + "","setang 90 0 0"])
      lines.append(["nav_end_area","nav_toggle_in_selected_set","mark FINALE","nav_clear_selected_set","clear_attribute FINALE"])
    
    walkables = self.map.root.FindRecurse(lambda node : "classname" in node.properties and node.properties["classname"] == "info_null" and "targetname" in node.properties and node.properties["targetname"] == "walkable")
    for walkable in walkables:
      lines.append(["setpos " + walkable.GetOrigin() + "","setang 90 0 0"])
      lines.append(["nav_mark_walkable"])
    lines.append(["nav_generate_incremental"])
    
    lines.append(["nav_analyze"])
    lines.append(["nav_save"])
    lines.append(["director_start","sv_cheats 0"])
    
    out = "bind KP_PLUS navgen000\n"
    for num, line in enumerate(lines):
      out += "alias \"navgen%03i\" \"%s;bind KP_PLUS navgen%03i\"\n"%(num,";".join(line),num+1)
    out += "alias \"navgen%03i\" \"echo Finished\"\n"%len(lines)
    return out