import numpy
import copy

def indent(depth):
  out = ""
  for d in range(depth):
    out += "  " 
  return out
  
def vectorToString(vector):
  out = ""
  for element in vector:
    out += "%i " % element # TODO: should be %e for precision
  return out[:-1]
  
def getBounds(points):
  return numpy.array([numpy.min(points, axis=0),numpy.max(points, axis=0)])

class VMFNode:  
  def __init__(self, name):
    self.name = name # structure in memory could be a dict tree (with names in dict keys)
    self.children = []
    self.properties = {}
    # TODO: save these as tuples
    self.plane = None
    self.origin = None
    
  def deepcopy(self):
    deepcopy = VMFNode(self.name)
    deepcopy.properties = copy.deepcopy(self.properties)
    if not self.plane == None:
      deepcopy.plane = self.plane.copy()
    else:
      deepcopy.plane = None
    if not self.origin == None:
      deepcopy.origin = self.origin.copy()
    else:
      deepcopy.origin = None
    for child in self.children:
      deepcopy.AddChild(child.deepcopy())
    return deepcopy
    
  # TODO: rename to append, top level Node may be python list
  def AddChild(self,child):
    self.children.append(child)
    
  def AddProperty(self,key,value):
    if key == "plane":
      self.SetPlane(value)
    elif key == "origin":
      self.SetOrigin(value)
    else:
      self.properties[key] = value
      
  def translatePlane(self,vector):
    for index, corner in enumerate(self.plane):
      self.plane[index] = corner + vector
    
  def translateOrigin(self,vector):
    self.origin = self.origin + vector
    if "BasisOrigin" in self.properties:
      basisOrigin = numpy.int_(numpy.fromstring(self.properties["BasisOrigin"], dtype=float, sep=' ').reshape(3)) # TODO: remove int conversion for precision - neccessary?
      basisOrigin = basisOrigin + vector
      self.properties["BasisOrigin"] = vectorToString(basisOrigin)
  
  def SetOrigin(self,origin):
    self.origin = numpy.int_(numpy.fromstring(origin, dtype=float, sep=' ').reshape(3)) # TODO: remove int conversion for precision - neccessary?
    
  def GetOrigin(self):
    return vectorToString(self.origin)
      
  def SetPlane(self,plane):
    plane = plane.translate(None,"()")
    self.plane = numpy.int_(numpy.fromstring(plane, dtype=float, sep=' ').reshape((3,3))) # TODO: type should be int for clean maps anyway
    
  def GetPlane(self):
    out = ""
    for corner in self.plane:
      out += "(" + vectorToString(corner) + ") "
    return out[:-1]
  
  def TranslateRecurse(self,vector):
    if not self.origin == None:
      self.translateOrigin(vector)
    elif not self.plane == None:
      self.translatePlane(vector)
    for child in self.children:
      child = child.TranslateRecurse(vector)
    return self
      
  def ToStringRecurse(self,depth):
    if not self.name == None:
      output = indent(depth) + self.name + "\n" + indent(depth) +"{\n"
      for key, value in self.properties.items():
        output += indent(depth+1) + "\""+key+"\" \""+value+"\"\n"
      if not self.origin == None:
        output += indent(depth+1) + "\"origin\" \""+self.GetOrigin()+"\"\n"
      if not self.plane == None:
        output += indent(depth+1) + "\"plane\" \""+self.GetPlane()+"\"\n"
    else:
      output = ""
      depth -= 1
    for child in self.children:
      output += child.ToStringRecurse(depth+1)
    if not self.name == None:
      output += indent(depth) + "}\n"
    return output
    
  def IncreaseIdRecurse(self,increase):
    if "id" in self.properties:
      self.properties["id"] = str(int(self.properties["id"]) + increase)
    if "sides" in self.properties:
      sides = self.properties["sides"].split(" ")
      self.properties["sides"] = ""
      for side in sides:
        self.properties["sides"] += str(int(side) + increase)
    for child in self.children:
      child = child.IncreaseIdRecurse(increase)
    return self
  
  def GetMaximumIdRecurse(self,maxId):
    if "id" in self.properties:
      id = int(self.properties["id"])
      if id > maxId:
        maxId = id
    for child in self.children:
      id = child.GetMaximumIdRecurse(maxId)
      if id > maxId:
        maxId = id
    return maxId
  
  def getWorldIndex(self):
    out = None
    for index in range(len(self.children)):
      if self.children[index].name == "world":
        out = index
    return out
  
  def AddOtherMap(self,otherMap):
    worldIndex = self.getWorldIndex()
    world = self.children[worldIndex]
    for otherNode in otherMap.children:
      if otherNode.name == "world":
        for child in otherNode.children:
          world.AddChild(child)
      elif otherNode.name == "entity":
        self.AddChild(otherNode)
    self.children[worldIndex] = world # index still valid because of append in AddChild
  
  def FindRecurse(self,predicate):
    hits = []
    if predicate(self):
      hits = [self]
    else:
      for child in self.children:
        hits.extend(child.FindRecurse(predicate))
    return hits
  
  def DeleteRecurse(self,predicate):
    removed = 0
    for child in self.children:
      if predicate(child):
        self.children.remove(child)
        removed += 1
      else:
        removed += child.DeleteRecurse(predicate)
    return removed
    
  def GetBoundsRecurse(self):
    bounds = None
    if not self.plane == None:
      bounds = getBounds(self.plane)
    for child in self.children:
      childBounds = child.GetBoundsRecurse()
      if not childBounds == None:
        if bounds == None:
          bounds = childBounds
        else:
          combined = numpy.append(childBounds,bounds).reshape((4,3))
          bounds = getBounds(combined)
    return bounds
  
  def __str__(self): 
    return "VMFNode type "+self.name